import json
from collections import namedtuple
from datetime import datetime
from pathlib import Path
from zipfile import ZipFile

import h5py
import numpy as np
from scipy.signal import convolve2d
from torch.utils.data import Dataset

from .exceptions import InconsistentDataException, DoesNotExistException
from .transforms import (
    DataTransform,
    MovieTransform,
    StaticTransform,
    Invertible,
    Subsequence,
    Delay,
)
from .utils import convert_static_h5_dataset_to_folder, zip_dir
from ..utils import recursively_load_dict_contents_from_group


class AttributeHandler:
    def __init__(self, name, h5_handle):
        """
        Can be used to turn a dataset within a hdf5 dataset into an attribute.

        Args:
            name:       name of the dataset in the hdf5 file
            h5_handle:  file handle for the hdf5 file
        """
        assert name in h5_handle, "{} must be in {}".format(name, h5_handle)
        self.name = name
        self.h5_handle = h5_handle

    def __getattr__(self, item):
        if item in self.h5_handle[self.name]:
            ret = self.h5_handle[self.name][item][()]
            if ret.dtype.char == "S":  # convert bytes to unicode
                ret = ret.astype(str)
            return ret
        else:
            raise AttributeError("Attribute {} not found".format(item))

    def __getitem__(self, item):
        return getattr(self, item)

    def keys(self):
        return self.h5_handle[self.name].keys()

    def __dir__(self):
        attrs = set(super().__dir__())
        return attrs.union(set(self.h5_handle[self.name].keys()))


class AttributeTransformer(AttributeHandler):
    def __init__(self, name, h5_handle, transforms, data_group):
        """
        Allows for id_transform of transforms to be applied to the
        specified attribute. Otherwise behaves like an AttributeHandler

        Args:
            name:       see AttributeHandler
            h5_handle:  see AttributeHandler
            transforms: the set of transforms that's supposed to be applied
            data_group: the data_key of the dataset that this attribute represents
        """
        super().__init__(name, h5_handle)
        self.transforms = transforms
        self.data_group = data_group

    def __getattr__(self, item):
        ret = {self.data_group: super().__getattr__(item)}
        for tr in self.transforms:
            if hasattr(tr, "id_transform"):
                ret = tr.id_transform(ret)

        return ret[self.data_group]


class TransformDataset(Dataset):
    def __init__(self, transforms=None):
        """
        Abstract Class for Datasets with transformations, providing `transform` and `invert` functions
        to apply data transformation on the elements.

        Args:
            transforms: list of transforms to be applied to each data point
        """
        self.transforms = transforms or []

    def transform(self, x, exclude=None):
        """
        Apply transform on a data element from the dataset

        Args:
            x (tuple): a data element from the dataset
            exclude (Transform, optional): Type of data transformer to be excluded from transform list. Defaults to None.

        Returns:
            tuple: transformed data element
        """

        for tr in self.transforms:
            if exclude is None or not isinstance(tr, exclude):
                x = tr(x)
        return x

    def invert(self, x, exclude=None):
        for tr in reversed(
            filter(lambda tr: not isinstance(tr, exclude), self.transforms)
        ):
            if not isinstance(tr, Invertible):
                raise TypeError("Cannot invert", tr.__class__.__name__)
            else:
                x = tr.inv(x)
        return x

    def __iter__(self):
        yield from map(self.__getitem__, range(len(self)))

    def __repr__(self):
        return (
            "{} m={}:\n\t({})".format(
                self.__class__.__name__, len(self), ", ".join(self.data_groups)
            )
            + "\n\t[Transforms: "
            + "->".join([repr(tr) for tr in self.transforms])
            + "]"
        )


class H5SequenceSet(TransformDataset):
    def __init__(self, filename, *data_keys, output_rename=None, transforms=None):
        super().__init__(transforms=transforms)

        if output_rename is None:
            output_rename = {}

        # a flag that can be changed to turn renaming on/off
        self.rename_output = True

        self.output_rename = output_rename

        self._fid = h5py.File(filename, "r")
        self.data = self._fid
        self.data_loaded = False

        # ensure that all elements of
        m = None
        for key in data_keys:
            assert key in self.data, "Could not find {} in file".format(key)
            l = len(self.data[key])
            if m is not None and l != m:
                raise ValueError("groups have different length")
            m = l
        self._len = m

        # Specify which types of transforms are accepted
        self._transform_set = DataTransform

        self.data_keys = data_keys
        self.transforms = transforms or []

        self.data_point = namedtuple("DataPoint", data_keys)
        self.output_point = namedtuple(
            "OutputPoint", [output_rename.get(k, k) for k in data_keys]
        )

    def load_content(self):
        self.data = recursively_load_dict_contents_from_group(self._fid)
        self.data_loaded = True

    def unload_content(self):
        self.data = self._fid
        self.data_loaded = False

    def __len__(self):
        return self._len

    def __getitem__(self, item):
        x = self.data_point(
            *(
                np.array(self.data[g][item if self.data_loaded else str(item)])
                for g in self.data_keys
            )
        )
        for tr in self.transforms:
            assert isinstance(tr, self._transform_set)
            x = tr(x)

        # convert to output point
        if self.rename_output:
            x = self.output_point(*x)
        return x

    def __getattr__(self, item):
        if item in self.data:
            item = self.data[item]
            if isinstance(item, h5py.Dataset):
                dtype = item.dtype
                item = item[()]
                if dtype.char == "S":  # convert bytes to unicode
                    item = item.astype(str)
                return item
            return item
        else:
            # TODO: check for a proper way to handle cases where super doesn't have __getattr__
            return super().__getattr__(item)

    def __repr__(self):
        names = [
            "{} -> {}".format(k, self.output_rename[k])
            if k in self.output_rename
            else k
            for k in self.data_keys
        ]
        s = "{} m={}:\n\t({})".format(
            self.__class__.__name__, len(self), ", ".join(names)
        )
        if self.transforms is not None:
            s += (
                "\n\t[Transforms: "
                + "->".join([repr(tr) for tr in self.transforms])
                + "]"
            )
        return s


class MovieSet(H5SequenceSet):
    """
    Extension to H5SequenceSet with specific HDF5 dataset assumed. Specifically,
    it assumes that properties such as `neurons` and `stats` are present in the dataset.
    """

    def __init__(
        self,
        filename,
        *data_groups,
        output_rename=None,
        transforms=None,
        stats_source="all"
    ):
        super().__init__(
            filename, *data_groups, output_rename=output_rename, transforms=transforms
        )
        self.stats_source = stats_source

        # set to accept only MovieTransform
        self._transform_set = MovieTransform

    @property
    def neurons(self):
        return AttributeTransformer(
            "neurons", self.data, self.transforms, data_group="responses"
        )

    @property
    def n_neurons(self):
        return len(self.neurons.unit_ids)

    @property
    def input_shape(self):
        name = (
            self.output_rename.get("inputs", "inputs")
            if self.rename_output
            else "inputs"
        )
        return (1,) + getattr(self[0], name).shape

    def transformed_mean(self, stats_source=None):
        if stats_source is None:
            stats_source = self.stats_source

        tmp = [
            np.atleast_1d(self.statistics[g][stats_source]["mean"][()])
            for g in self.data_keys
        ]
        x = self.transform(self.data_point(*tmp), exclude=(Subsequence, Delay))
        if self.rename_output:
            x = self.output_point(*x)
        return x

    def rf_base(self, stats_source="all"):
        N, c, t, w, h = self.img_shape
        t = min(t, 150)
        mean = lambda dk: self.statistics[dk][stats_source]["mean"][()]
        d = dict(
            inputs=np.ones((1, c, t, w, h)) * np.array(mean("inputs")),
            eye_position=np.ones((1, t, 1)) * mean("eye_position")[None, None, :],
            behavior=np.ones((1, t, 1)) * mean("behavior")[None, None, :],
            responses=np.ones((1, t, 1)) * mean("responses")[None, None, :],
        )
        return self.transform(
            self.data_point(*[d[dk] for dk in self.data_keys]), exclude=Subsequence
        )

    def rf_noise_stim(self, m, t, stats_source="all"):
        """
        Generates a Gaussian white noise stimulus filtered with a 3x3 Gaussian filter
        for the computation of receptive fields. The mean and variance of the Gaussian
        noise are set to the mean and variance of the stimulus ensemble.

        The behvavior, eye movement statistics, and responses are set to their respective means.
        Args:
            m: number of noise samples
            t: length in time

        Returns: tuple of input, behavior, eye, and response

        """
        N, c, _, w, h = self.img_shape
        stat = lambda dk, what: self.statistics[dk][stats_source][what][()]
        mu, s = stat("inputs", "mean"), stat("inputs", "std")
        h_filt = np.float64(
            [[1 / 16, 1 / 8, 1 / 16], [1 / 8, 1 / 4, 1 / 8], [1 / 16, 1 / 8, 1 / 16]]
        )
        noise_input = (
            np.stack(
                [
                    convolve2d(np.random.randn(w, h), h_filt, mode="same")
                    for _ in range(m * t * c)
                ]
            ).reshape((m, c, t, w, h))
            * s
            + mu
        )

        mean_beh = np.ones((m, t, 1)) * stat("behavior", "mean")[None, None, :]
        mean_eye = np.ones((m, t, 1)) * stat("eye_position", "mean")[None, None, :]
        mean_resp = np.ones((m, t, 1)) * stat("responses", "mean")[None, None, :]

        d = dict(
            inputs=noise_input.astype(np.float32),
            eye_position=mean_eye.astype(np.float32),
            behavior=mean_beh.astype(np.float32),
            responses=mean_resp.astype(np.float32),
        )

        return self.transform(
            self.data_point(*[d[dk] for dk in self.data_groups.values()]),
            exclude=(Subsequence, Delay),
        )


default_datapoint = namedtuple("DefaultDataPoint", ["images", "responses"])


class StaticSet(TransformDataset):
    def __init__(self, *data_keys, transforms=None):
        """
        Abstract class for static datasets. Defines data_keys and a corresponding datapoint.
        """
        super().__init__(transforms=transforms)

        self.data_keys = data_keys
        if set(data_keys) == {"images", "responses"}:
            # this version IS serializable in pickle
            self.data_point = default_datapoint
        else:
            # this version is NOT - you cannot use this with a dataloader with num_workers > 1
            self.data_point = namedtuple("DataPoint", data_keys)


class H5ArraySet(StaticSet):
    def __init__(self, filename, *data_keys, transforms=None):
        """
        Dataset for static data stored in hdf5 files.

        Args:
            filename:      filename of the hdf5 file
            *data_keys:    data keys to be read from the file
            transforms:    list of transforms applied to each datapoint
        """
        super().__init__(*data_keys, transforms=transforms)

        self._fid = h5py.File(filename, "r")
        self.data = self._fid
        self.data_loaded = False
        m = None
        for key in data_keys:
            assert key in self.data, "Could not find {} in file".format(key)
            if m is None:
                m = len(self.data[key])
            else:
                assert m == len(self.data[key]), "Length of datasets do not match"
        self._len = m

    def load_content(self):
        self.data = recursively_load_dict_contents_from_group(self._fid)
        self.data_loaded = True

    def unload_content(self):
        self.data = self._fid
        self.data_loaded = False

    def __getitem__(self, item):
        x = self.data_point(*(self.data[g][item] for g in self.data_keys))
        for tr in self.transforms:
            assert isinstance(tr, StaticTransform)
            x = tr(x)
        return x

    def __iter__(self):
        yield from map(self.__getitem__, range(len(self)))

    def __len__(self):
        return self._len

    def __repr__(self):
        return "\n".join(
            [
                "Tensor {}: {} ".format(key, self.data[key].shape)
                for key in self.data_keys
            ]
            + ["Transforms: " + repr(self.transforms)]
        )

    def __getattr__(self, item):
        if item in self.data:
            item = self.data[item]
            if isinstance(item, h5py.Dataset):
                dtype = item.dtype
                item = item[()]
                if dtype.char == "S":  # convert bytes to univcode
                    item = item.astype(str)
                return item
            return item
        else:
            raise AttributeError(
                "Item {} not found in {}".format(item, self.__class__.__name__)
            )


class StaticImageSet(H5ArraySet):
    def __init__(
        self, filename, *data_keys, transforms=None, cache_raw=False, stats_source=None
    ):
        """
        Dataset for h5 files.

        Args:
            filename:       filename of the hdf5 file
            *data_keys:     datasets to be extracted
            transforms:     transforms applied to each data point
            cache_raw:      whether to cache the raw (untransformed) datapoints
            stats_source:   statistic source to be used.
        """
        super().__init__(filename, *data_keys, transforms=transforms)
        self.cache_raw = cache_raw
        self.last_raw = None
        self.stats_source = stats_source if stats_source is not None else "all"

    @property
    def n_neurons(self):
        return len(self[0].responses)

    @property
    def neurons(self):
        return AttributeTransformer(
            "neurons", self.data, self.transforms, data_group="responses"
        )

    @property
    def info(self):
        return AttributeHandler("item_info", self.data)

    @property
    def img_shape(self):
        return (1,) + self[0].images.shape

    def transformed_mean(self, stats_source=None):
        if stats_source is None:
            stats_source = self.stats_source

        tmp = [
            np.atleast_1d(self.statistics[dk][stats_source]["mean"][()])
            for dk in self.data_keys
        ]
        return self.transform(self.data_point(*tmp))

    def __repr__(self):
        return super().__repr__() + (
            "\n\t[Stats source: {}]".format(self.stats_source)
            if self.stats_source is not None
            else ""
        )

    def __dir__(self):
        attrs = set(self.__dict__).union(set(dir(type(self))))
        return attrs.union(set(self.data.keys()))


class DirectoryAttributeHandler:
    def __init__(self, path, links=None):
        """
        Class that can be used to represent a subdirectory of a FileTree as a property in a FileTree dataset.
        Caches already loaded data items.

        Args:
            path: path to the subdiretory (pathlib.Path object)
        """
        self.links = links or {}
        self.path = path

    def __getattr__(self, item):
        temp_path = self.resolve_item_path(item)
        if temp_path.exists() and temp_path.is_dir():
            val = DirectoryAttributeHandler(temp_path, links=self.links)
        else:
            val = np.load(self.path / "{}.npy".format(item))
        return val

    def resolve_item_path(self, item):
        if item in self.links:
            item = self.links[item]
        return self.path / item

    def __getitem__(self, item):
        return getattr(self, item)

    def keys(self):
        return [e.stem for e in self.path.glob("*")]

    def __dir__(self):
        attrs = set(super().__dir__())
        return attrs.union(set(self.keys())).union(set(self.links.keys()))


class DirectoryAttributeTransformer(DirectoryAttributeHandler):
    def __init__(self, path, transforms, data_group, links=None):
        """
        Class that can be used to represent a subdirectory of a FileTree as a property in a FileTree dataset.
        Like DirectoryAttributeHandler but allows for id_transform of transforms to be applied to the
        specified attribute.

        Args:
            path: path to the subdiretory (pathlib.Path object)
        """

        super().__init__(path, links=links)
        self.transforms = transforms
        self.data_group = data_group

    def __getattr__(self, item):
        ret = {self.data_group: super().__getattr__(item)}
        for tr in self.transforms:
            ret = tr.id_transform(ret)
        return ret[self.data_group]


class FileTreeDataset(StaticSet):
    def __init__(self, dirname, *data_keys, transforms=None):
        """
        Dataset stored as a file tree. The tree needs to have the subdirs data, meta, meta/neurons, meta/statistics,
        and meta/trials. Please refer to convert_static_h5_dataset_to_folder in mlutils.data.utils
        how to export an hdf5 file into that structure.


        Here is an example. Data directories with too many entries have trials as .npy files
        named 0.npy, 1.npy, ...
        The meta/trials subdirectory must have single .npy files with arrays that provide additional trial based
        meta data.

        static22564-2-13-preproc0
        ├── data
        │   ├── behavior [5955 entries exceeds filelimit, not opening dir]
        │   ├── images [5955 entries exceeds filelimit, not opening dir]
        │   ├── pupil_center [5955 entries exceeds filelimit, not opening dir]
        │   └── responses [5955 entries exceeds filelimit, not opening dir]
        └── meta
            ├── neurons
            │   ├── animal_ids.npy
            │   ├── area.npy
            │   ├── layer.npy
            │   ├── scan_idx.npy
            │   ├── sessions.npy
            │   └── unit_ids.npy
            ├── statistics
            │   ├── behavior
            │   │   ├── all
            │   │   │   ├── max.npy
            │   │   │   ├── mean.npy
            │   │   │   ├── median.npy
            │   │   │   ├── min.npy
            │   │   │   └── std.npy
            │   │   └── stimulus_frame
            │   │       ├── max.npy
            │   │       ├── mean.npy
            │   │       ├── median.npy
            │   │       ├── min.npy
            │   │       └── std.npy
            │   ├── images
            │   │   ├── all
            │   │   │   ├── max.npy
            │   │   │   ├── mean.npy
            │   │   │   ├── median.npy
            │   │   │   ├── min.npy
            │   │   │   └── std.npy
            │   │   └── stimulus_frame
            │   │       ├── max.npy
            │   │       ├── mean.npy
            │   │       ├── median.npy
            │   │       ├── min.npy
            │   │       └── std.npy
            │   ├── pupil_center
            │   │   ├── all
            │   │   │   ├── max.npy
            │   │   │   ├── mean.npy
            │   │   │   ├── median.npy
            │   │   │   ├── min.npy
            │   │   │   └── std.npy
            │   │   └── stimulus_frame
            │   │       ├── max.npy
            │   │       ├── mean.npy
            │   │       ├── median.npy
            │   │       ├── min.npy
            │   │       └── std.npy
            │   └── responses
            │       ├── all
            │       │   ├── max.npy
            │       │   ├── mean.npy
            │       │   ├── median.npy
            │       │   ├── min.npy
            │       │   └── std.npy
            │       └── stimulus_frame
            │           ├── max.npy
            │           ├── mean.npy
            │           ├── median.npy
            │           ├── min.npy
            │           └── std.npy
            └── trials [12 entries exceeds filelimit, not opening dir]

        Args:
            dirname:     root directory name
            *data_keys:  data items to be extraced (must be subdirectories of root/data)
            transforms:  transforms to be applied to the data (see TransformDataset)
        """
        super().__init__(*data_keys, transforms=transforms)

        number_of_files = []

        if dirname.endswith(".zip"):
            if not Path(dirname[:-4]).exists():
                self.unzip(dirname, Path(dirname).absolute().parent)
            else:
                print(
                    "{} exists already. Not unpacking {}".format(dirname[:-4], dirname)
                )
            dirname = dirname[:-4]

        self.basepath = Path(dirname).absolute()
        self._config_file = self.basepath / "config.json"

        if not self._config_file.exists():
            self._save_config(self._default_config)

        for data_key in data_keys:
            datapath = self.resolve_data_path(data_key)
            number_of_files.append(len(list(datapath.glob("*"))))

        if not np.all(np.diff(number_of_files) == 0):
            raise InconsistentDataException("Number of data points is not equal")
        else:
            self._len = number_of_files[0]

        self._cache = {data_key: {} for data_key in data_keys}

    _default_config = {"links": {}}

    def resolve_data_path(self, data_key):
        if self.link_exists(data_key):
            data_key = self.config["links"][data_key]
        datapath = self.basepath / "data" / data_key

        if not datapath.exists():
            raise DoesNotExistException("Data path {} does not exist".format(datapath))
        return datapath

    def link_exists(self, link):
        return "links" in self.config and link in self.config["links"]

    @property
    def config(self):
        with open(self._config_file) as fid:
            return json.load(fid)

    def _save_config(self, cfg):
        with open(self._config_file, "w") as fid:
            return json.dump(cfg, fid)

    def __len__(self):
        return self._len

    def __getitem__(self, item):
        # load data from cache or disk
        ret = []
        for data_key in self.data_keys:
            if item in self._cache[data_key]:
                ret.append(self._cache[data_key][item])
            else:
                datapath = self.resolve_data_path(data_key)
                val = np.load(datapath / "{}.npy".format(item))
                self._cache[data_key][item] = val
                ret.append(val)

        # create data point and transform
        x = self.data_point(*ret)
        for tr in self.transforms:
            assert isinstance(tr, StaticTransform)
            x = tr(x)
        return x

    def add_log_entry(self, msg):
        timestamp = datetime.now().strftime("%d-%b-%Y (%H:%M:%S.%f)")
        with open(self.basepath / "change.log", "a+") as fid:
            fid.write("{}: {}\n".format(timestamp, msg))

    @staticmethod
    def match_order(target, permuted, not_exist_ok=False):
        """
        Matches the order or rows in permuted to by returning an index array such that.

        Args:
            not_exist_ok: if the element does not exist, don't return an index

        Returns: index array `idx` such that `target == permuted[idx, :]`
        """

        order, target_idx = [], []
        unmatched_counter = 0
        for i, row in enumerate(target):
            idx = np.sum(permuted - row, axis=1) == 0
            if not not_exist_ok:
                assert idx.sum() == 1
            if idx.sum() == 1:
                order.append(np.where(idx)[0][0])
                target_idx.append(i)
            else:
                unmatched_counter += 1
        if not_exist_ok:
            print("Encountered {} unmatched elements".format(unmatched_counter))
        return np.array(target_idx, dtype=int), np.array(order, dtype=int)

    def add_neuron_meta(
        self, name, animal_id, session, scan_idx, unit_id, values, fill_missing=None
    ):
        """
        Add new meta information about neurons.

        Args:
            name:       name of the new meta information
            animal_id:  array with animal_ids per first dimension of values
            session:    array with session per first dimension of values
            scan_idx:   array with scan_idx per first dimension of values
            unit_id:    array with unit_id per first dimension of values
            values:     new meta information. First dimension must refer to neurons.
            fill_missing:   fill the values of the new attribute with NaN if not provided
        """
        if (
            not len(animal_id)
            == len(session)
            == len(scan_idx)
            == len(unit_id)
            == len(values)
        ):
            raise InconsistentDataException(
                "number of trials and identifiers not consistent"
            )

        target = np.c_[
            (
                self.neurons.animal_ids,
                self.neurons.sessions,
                self.neurons.scan_idx,
                self.neurons.unit_ids,
            )
        ]
        permuted = np.c_[(animal_id, session, scan_idx, unit_id)]
        vals = np.ones((len(target),) + values.shape[1:], dtype=values.dtype) * (
            np.nan if fill_missing is None else fill_missing
        )
        tidx, idx = self.match_order(
            target, permuted, not_exist_ok=fill_missing is not None
        )

        assert (
            np.sum(target[tidx] - permuted[idx, ...]) == 0
        ), "Something went wrong in sorting"

        vals[tidx, ...] = values[idx, ...]
        np.save(self.basepath / "meta/neurons/{}.npy".format(name), vals)
        self.add_log_entry(
            "Added new neuron meta attribute {} to meta/neurons".format(name)
        )

    @staticmethod
    def initialize_from(filename, outpath=None, overwrite=False):
        """
        Convenience function. See `convert_static_h5_dataset_to_folder` in `.utils`
        """
        convert_static_h5_dataset_to_folder(
            filename, outpath=outpath, overwrite=overwrite
        )

    @property
    def change_log(self):
        if (self.basepath / "change.log").exists():
            with open(self.basepath / "change.log", "r") as fid:
                print("".join(fid.readlines()))

    def zip(self, filename=None):
        """
        Zips current dataset.

        Args:
            filename:  Filename for the zip. Directory name + zip by default.
        """

        if filename is None:
            filename = str(self.basepath) + ".zip"
        zip_dir(filename, self.basepath)

    def unzip(self, filename, path):
        print("Unzipping {} into {}".format(filename, path))
        with ZipFile(filename, "r") as zip_obj:
            zip_obj.extractall(path)

    def add_link(self, attr, new_name):
        """
        Add a new dataset that links to an existing dataset.

        For instance `targets` that links to `responses`

        Args:
            attr:       existing attribute such as `responses`
            new_name:   name of the new attribute reference.
        """
        if not (self.basepath / "data/{}".format(attr)).exists():
            raise DoesNotExistException("Link target does not exist")

        if (self.basepath / "data/{}".format(new_name)).exists():
            raise FileExistsError("Link target already exists")

        config = self.config
        if not "links" in config:
            config["links"] = {}
        config["links"][new_name] = attr
        self._save_config(config)

    @property
    def n_neurons(self):
        return len(self[0].responses)

    @property
    def neurons(self):
        return DirectoryAttributeTransformer(
            self.basepath / "meta/neurons",
            self.transforms,
            data_group="responses" if "responses" in self.data_keys else "targets",
        )

    @property
    def trial_info(self):
        return DirectoryAttributeHandler(self.basepath / "meta/trials")

    @property
    def statistics(self):
        return DirectoryAttributeHandler(
            self.basepath / "meta/statistics", self.config["links"]
        )

    @property
    def img_shape(self):
        return (1,) + self[0].images.shape

    def __repr__(self):
        return "{} {} (n={} items)\n\t{}".format(
            self.__class__.__name__, self.basepath, self._len, ", ".join(self.data_keys)
        )



import numpy as np
import torch
from collections import namedtuple


class Invertible:
    def inv(self, y):
        raise NotImplemented("Subclasses of Invertible must implement an inv method")


class DataTransform:
    def __repr__(self):
        return self.__class__.__name__

    def id_transform(self, id_map):
        """
        Given a dictionary mapping from data group name to
        a numpy array of appropriate size containing identity information
        for that particular group name, the transform is expected to return
        the expected identity modification (e.g. dropping or duplication of entries, reordering, etc)
        """
        return id_map

    def __call__(self, *args, **kwargs):
        raise NotImplementedError


class MovieTransform(DataTransform):
    """
    Abstract class to certify that the transform is valid for sequence like (e.g. movie) datasets.
    """


class StaticTransform(DataTransform):
    """
    Abstract class to certify that the transform is valid for non-sequential (e.g. image) datasets.
    """


class Subsequence(MovieTransform):
    def __init__(self, frames, channel_first=("inputs",), offset=-1):
        """
        Given a sequential (movie like) data, subselect a consequent `frames` counts of frames, starting with
        `offset` frames skipped. If `offset`< 0, then the subsequence is taken with a random (but valid) offset each iteration.

        Args:
            frames (int): Length of subsequence to be selected from each sample
            channel_first (tuple, optional): A list of data key names where the channel (and thus not time) dimension occurs on the first dimension (dim=0). Otherwise, it's assumed
            that the time dimesion occurs on the first dimension. Defaults to ("inputs",).
            offset (int, optional): Offset to start the subsequence from. Defaults to -1, corresponding to random but valid offset at each iteration.
        """
        self.frames = frames
        # for special group, the time slicing is applied on
        self.channel_first = channel_first
        self.offset = offset

    def __call__(self, x):
        first_group = x._fields[0]

        # get the length of the full sequence. Note that which axis to pick depends on if it's
        # channel first (i.e. time second) group
        t = getattr(x, first_group).shape[int(first_group in self.channel_first)]

        if self.offset < 0:
            i = np.random.randint(0, t - self.frames)
        else:
            i = self.offset
        return x.__class__(
            **{
                k: getattr(x, k)[:, i : i + self.frames, ...]
                if k in self.channel_first
                else getattr(x, k)[i : i + self.frames, ...]
                for k in x._fields
            }
        )

    def id_transform(self, id_map):
        # until a better solution is reached, skipping this
        return id_map

        new_map = {}
        first_group = list(id_map.keys())[0]
        v_fg = id_map[first_group]
        t = v_fg.shape[int(first_group in self.channel_first)]
        i = np.random.randint(0, t - self.frames)
        for k, v in id_map.items():
            if k in self.channel_first:
                new_map[k] = v[:, i : i + self.frames]
            else:
                new_map[k] = v[i : i + self.frames]

        return new_map

    def __repr__(self):
        return self.__class__.__name__ + "({})".format(self.frames)


class Delay(MovieTransform):
    """
    Delay the specified target gorups by delay frames. In other words,
    given non-delayed group g(t) and delayed group d(t), 
    g(T:N-delay) will be returned with d(T+delay:N) where `delay` is the specified amount
    of delay, and N is the last frame in the dataset.
    """

    def __init__(self, delay, delay_groups=("responses",), channel_first=("inputs",)):
        self.delay = delay
        self.delay_groups = delay_groups
        self.channel_first = channel_first

    def __call__(self, x):
        first_group = x._fields[0]
        t = getattr(x, first_group).shape[int(first_group in self.channel_first)]
        assert (
            t > self.delay
        ), "The sequence length {} has to be longer than the delay {}".format(
            t, self.delay
        )
        key_entry = {}
        for k in x._fields:
            if k in self.delay_groups:
                start, stop = self.delay, t
            else:
                start = 0
                stop = t - self.delay

            key_entry[k] = (
                getattr(x, k)[:, start:stop]
                if k in self.channel_first
                else getattr(x, k)[start:stop]
            )

        return x.__class__(**key_entry)

    def id_transform(self, id_map):
        # until a better solution is reached, skipping this
        return id_map

        new_map = {}
        first_group = list(id_map.keys())[0]
        v_fg = id_map[first_group]
        t = v_fg.shape[int(first_group in self.channel_first)]
        assert (
            t > self.delay
        ), "The sequence length {} has to be longer than the delay {}".format(
            t, self.delay
        )

        for k, v in id_map.items():
            if k in self.delay_groups:
                start, end = self.delay, t
            else:
                start, end = 0, t - self.delay

            new_map[k] = v[:, start:end] if k in self.channel_first else v[start:end]

        return new_map

    def __repr__(self):
        return self.__class__.__name__ + "({} on {})".format(
            self.delay, self.delay_groups
        )


class Subsample(MovieTransform, StaticTransform):
    def __init__(self, idx, target_group="responses", target_index=None):
        self.idx = idx  # torch.tensor(idx, dtype=torch.long)
        if isinstance(target_group, str):
            self.target_groups = (target_group,)
        else:
            self.target_groups = target_group

        if target_index is None:
            target_index = {k: -1 for k in self.target_groups}
        elif isinstance(target_index, int):
            target_index = {k: target_index for k in self.target_groups}
        self.target_index = target_index

        assert np.ndim(self.idx) == 1, "Dimensionality of index array has to be 1"

    def __call__(self, x):
        return x.__class__(
            **{
                k: np.take(getattr(x, k), self.idx, self.target_index[k])
                if k in self.target_groups
                else getattr(x, k)
                for k in x._fields
            }
        )

    def __repr__(self):
        return self.__class__.__name__ + "(n={})".format(len(self.idx))

    def id_transform(self, id_map):
        return {
            k: v[self.idx] if k in self.target_groups else v for k, v in id_map.items()
        }


class ToTensor(MovieTransform, StaticTransform, Invertible):
    def __init__(self, cuda=False):
        self.cuda = cuda

    def inv(self, y):
        return y.numpy()

    def __call__(self, x):
        return x.__class__(
            *[
                torch.from_numpy(elem.astype(np.float32)).cuda()
                if self.cuda
                else torch.from_numpy(elem.astype(np.float32))
                for elem in x
            ]
        )


class Identity(MovieTransform, StaticTransform, Invertible):
    def __call__(self, x):
        return x

    def inv(self, y):
        return y


class Rename(MovieTransform, StaticTransform, Invertible):
    def __init__(self, name_map):
        self.name_map = name_map
        self.rev_map = {v: k for k, v in name_map.items()}
        self.tuple_class = None
        self.origin_tuple_class = None

    def __call__(self, x):
        if self.tuple_class is None:
            if self.origin_tuple_class is None:
                self.origin_tuple_class = x.__class__
            renamed_fields = [self.name_map.get(f, f) for f in x._fields]
            self.tuple_class = namedtuple("RenamedDataPoint", renamed_fields)
        return self.tuple_class(*x)

    def inv(self, y):
        if self.origin_tuple_class is None:
            renamed_fields = [self.rev_map.get(f, f) for f in y._fields]
            self.origin_tuple_class = namedtuple("OriginalDataPoint", renamed_fields)
        return self.origin_tuple_class(*y)

    def id_transform(self, id_map):
        return {self.name_map.get(k, k): v for k, v in id_map.items()}


class NeuroNormalizer(MovieTransform, StaticTransform, Invertible):
    """
    Note that this normalizer only works with MovieDataset of very specific formulation

    Normalizes a trial with fields: inputs, behavior, eye_position, and responses. The pair of
    behavior and eye_position can be missing. The following normalizations are applied:

    - inputs are scaled by the training std of the stats_source and centered on the mean of the movie
    - behavior is divided by the std if the std is greater than 1% of the mean std (to avoid division by 0)
    - eye_position is z-scored
    - reponses are divided by the per neuron std if the std is greater than
            1% of the mean std (to avoid division by 0)
    """

    def __init__(self, data, stats_source="all", exclude=None):

        self.exclude = exclude or []

        in_name = "images" if "images" in data.statistics.keys() else "inputs"

        self._inputs_mean = data.statistics[in_name][stats_source]["mean"][()]
        self._inputs_std = data.statistics[in_name][stats_source]["std"][()]

        s = np.array(data.statistics["responses"][stats_source]["std"])

        # TODO: consider other baselines
        threshold = 0.01 * s.mean()
        idx = s > threshold
        self._response_precision = np.ones_like(s) / threshold
        self._response_precision[idx] = 1 / s[idx]
        transforms, itransforms = {}, {}

        # -- inputs
        transforms[in_name] = lambda x: (x - self._inputs_mean) / self._inputs_std
        itransforms[in_name] = lambda x: x * self._inputs_std + self._inputs_mean

        # -- responses
        transforms["responses"] = lambda x: x * self._response_precision
        itransforms["responses"] = lambda x: x / self._response_precision

        if "eye_position" in data.data_keys:
            # -- eye position
            self._eye_mean = np.array(
                data.statistics["eye_position"][stats_source]["mean"]
            )
            self._eye_std = np.array(
                data.statistics["eye_position"][stats_source]["std"]
            )
            transforms["eye_position"] = lambda x: (x - self._eye_mean) / self._eye_std
            itransforms["eye_position"] = lambda x: x * self._eye_std + self._eye_mean

            s = np.array(data.statistics["behavior"][stats_source]["std"])

            # TODO: same as above - consider other baselines
            threshold = 0.01 * s.mean()
            idx = s > threshold
            self._behavior_precision = np.ones_like(s) / threshold
            self._behavior_precision[idx] = 1 / s[idx]

            # -- behavior
            transforms["behavior"] = lambda x: x * self._behavior_precision
            itransforms["behavior"] = lambda x: x / self._behavior_precision

        self._transforms = transforms
        self._itransforms = itransforms

    def __call__(self, x):
        """
        Apply transformation
        """
        return x.__class__(
            **{
                k: (self._transforms[k](v) if k not in self.exclude else v)
                for k, v in zip(x._fields, x)
            }
        )

    def inv(self, x):
        return x.__class__(
            **{
                k: (self._itransforms[k](v) if k not in self.exclude else v)
                for k, v in zip(x._fields, x)
            }
        )

    def __repr__(self):
        return super().__repr__() + (
            "(not {})".format(", ".join(self.exclude))
            if self.exclude is not None
            else ""
        )


class AddBehaviorAsChannels(MovieTransform, StaticTransform, Invertible):
    """
    Given a StaticImage object that includes "images", "responses", and "behavior", it returns three variables:
        - input image concatinated with behavior as new channel(s)
        - responses
        - behavior
    """

    def __init__(self):
        self.transforms, self.itransforms = {}, {}
        self.transforms["images"] = lambda img, behavior: np.concatenate(
            (
                img,
                np.ones((1, *img.shape[-(len(img.shape) - 1) :]))
                * np.expand_dims(
                    behavior, axis=((len(img.shape) - 2), (len(img.shape) - 1))
                ),
            ),
            axis=len(img.shape) - 3,
        )
        self.transforms["responses"] = lambda x: x
        self.transforms["behavior"] = lambda x: x

    def __call__(self, x):
        key_vals = {k: v for k, v in zip(x._fields, x)}
        dd = {
            "images": self.transforms["images"](
                key_vals["images"], key_vals["behavior"]
            ),
            "responses": self.transforms["responses"](key_vals["responses"]),
            "behavior": self.transforms["behavior"](key_vals["behavior"]),
        }
        return x.__class__(**dd)


class SelectInputChannel(StaticTransform):
    """
    Given a StaticImage object that includes "images", it will select a particular input channel.
    """

    def __init__(self, grab_channel):
        self.grab_channel = grab_channel

    def __call__(self, x):
        key_vals = {k: v for k, v in zip(x._fields, x)}
        img = key_vals["images"]
        key_vals["images"] = (
            img[:, (self.grab_channel,)]
            if len(img.shape) == 4
            else img[(self.grab_channel,)]
        )
        return x.__class__(**key_vals)



#!/usr/bin/env python
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "requirements.txt")) as f:
    requirements = f.read().split()


setup(
    name="neuralpredictors",
    version="0.0.0rc1",
    description="Sinz Lab Neural System Identification Utilities",
    author="Sinz Lab",
    author_email="software@sinzlab.net",
    url="https://github.com/sinzlab/neuralpredictors",
    packages=find_packages(exclude=[]),
    install_requires=requirements,
)



import torch

from torch import nn


def constrain_all(self):
    if hasattr(self, "constrain"):
        self.constrain()

    for c in self.children():
        c.constrain_all()


# extend torch nn.Module to have constrain_all function
nn.Module.constrain_all = constrain_all


# all constrain function takes an optional cache argument
# the cache can be used to store a relatively expensive reusable
# item usable in the constraining. For example, the cache can store the
# binary map of all units that should be constrained.
def positive(weight):
    """
    Enforces tensor to be positive. Changes the tensor in place. Produces no gradient.

    Args:
        weight: tensor to be positive

    """
    # with torch.no_grad():
    #     weight.clamp_(0)
    weight.data *= weight.data.ge(0).float()


def at_least(weight, bound, cache=None):
    weight.data[weight < bound] = bound
    return cache


def at_most(weight, bound, cache=None):
    weight.data[weight > bound] = bound
    return cache


def negative(weight, cache=None):
    weight.data *= weight.data.le(0).float()
    return cache


def positive_except_self(weight, cache=None):
    pos = weight.data.ge(0).float()
    if pos.size()[2] % 2 == 0 or pos.size()[3] % 2 == 0:
        raise ValueError("kernel size must be odd")
    ii, jj = pos.size()[2] // 2, pos.size()[3] // 2
    for i in range(pos.size()[0]):
        pos[i, i, ii, jj] = 1
    weight.data *= pos
    return cache



import math
import numpy as np
from numpy import pi
from scipy.special import gamma
from numpy.polynomial.polynomial import polyval


import torch
from torch import nn


def hermite_coefficients(mu, nu):
    """Generate coefficients of 2D Hermite functions"""
    nur = np.arange(nu + 1)
    num = gamma(mu + nu + 1) * gamma(nu + 1) * ((-2) ** (nu - nur))
    denom = gamma(mu + 1 + nur) * gamma(1 + nur) * gamma(nu + 1 - nur)
    return num / denom


def hermite_2d(N, npts, xvalmax=None):
    """Generate 2D Hermite function basis

    Arguments:
    N           -- the maximum rank.
    npts        -- the number of points in x and y

    Keyword arguments:
    xvalmax     -- the maximum x and y value (default: 2.5 * sqrt(N))

    Returns:
    H           -- Basis set of size N*(N+1)/2 x npts x npts
    desc        -- List of descriptors specifying for each
                   basis function whether it is:
                        'z': rotationally symmetric
                        'r': real part of quadrature pair
                        'i': imaginary part of quadrature pair

    """
    xvalmax = xvalmax or 2.5 * np.sqrt(N)
    ranks = range(N)

    # Gaussian envelope
    xvalmax *= 1 - 1 / npts
    xvals = np.linspace(-xvalmax, xvalmax, npts, endpoint=True)[...,None]

    gxv = np.exp(-xvals ** 2 / 4)
    gaussian = np.dot(gxv, gxv.T)

    # Hermite polynomials
    mu = np.array([])
    nu = np.array([])
    desc = []
    for i, rank in enumerate(ranks):
        muadd = np.sort(np.abs(np.arange(-rank, rank + 0.1, 2)))
        mu = np.hstack([mu, muadd])
        nu = np.hstack([nu, (rank - muadd) / 2])
        if not (rank % 2):
            desc.append('z')
        desc += ['r', 'i'] * int(np.floor((rank + 1) / 2))

    theta = np.arctan2(xvals, xvals.T)
    radsq = xvals ** 2 + xvals.T ** 2
    nbases = mu.size
    H = np.zeros([nbases, npts, npts])
    for i, (mui, nui, desci) in enumerate(zip(mu, nu, desc)):
        radvals = polyval(radsq, hermite_coefficients(mui, nui))
        basis = gaussian * (radsq ** (mui / 2)) * radvals * np.exp(1j * mui * theta)
        basis /= np.sqrt(2 ** (mui + 2 * nui) * pi * \
                         math.factorial(mui + nui) * math.factorial(nui))
        if desci == 'z':
            H[i] = basis.real / np.sqrt(2)
        elif desci == 'r':
            H[i] = basis.real
        elif desci == 'i':
            H[i] = basis.imag

    # normalize
    return H / np.sqrt(np.sum(H ** 2, axis=(1, 2), keepdims=True)), desc, mu


def rotation_matrix(desc, mu, angle):
    R = np.zeros((len(desc), len(desc)))
    for i, (d, m) in enumerate(zip(desc, mu)):
        if d == 'r':
            Rc = np.array([[np.cos(m*angle), np.sin(m*angle)],
                           [-np.sin(m*angle), np.cos(m*angle)]])
            R[i:i+2,i:i+2] = Rc
        elif d == 'z':
            R[i,i] = 1
    return R


def downsample_weights(weights, factor=2):
    w = 0
    for i in range(factor):
        for j in range(factor):
            w += weights[i::factor,j::factor]
    return w


class RotateHermite(nn.Module):
    def __init__(self,
                 filter_size,
                 upsampling,
                 num_rotations,
                 first_layer):

        super().__init__()

        H, desc, mu = hermite_2d(
            filter_size, filter_size * upsampling, 2 * np.sqrt(filter_size))

        self.H = nn.Parameter(torch.tensor(H, dtype=torch.float32), requires_grad=False)

        angles = [i * 2 * pi / num_rotations for i in range(num_rotations)]
        Rs = [torch.tensor(rotation_matrix(desc, mu, angle), dtype=torch.float32)
              for angle in angles]

        self.Rs = nn.ParameterList([nn.Parameter(R, requires_grad=False) for R in Rs])

        self.num_rotations = num_rotations
        self.first_layer = first_layer

    def forward(self, coeffs):
        num_coeffs, num_inputs_total, num_outputs = coeffs.shape
        filter_size = self.H.shape[1]
        num_inputs = num_inputs_total // self.num_rotations
        weights_rotated = []
        for i, R in enumerate(self.Rs):
            coeffs_rotated = torch.tensordot(R, coeffs, dims=([1], [0]))
            w = torch.tensordot(self.H, coeffs_rotated, dims=[[0], [0]])
            if i and not self.first_layer:
                shift = num_inputs_total - i * num_inputs
                w = torch.cat([w[:,:,shift:,:], w[:,:,:shift,:]], dim=2)
            weights_rotated.append(w)
        weights_all_rotations = torch.cat(weights_rotated, dim=3)
        return weights_all_rotations


class HermiteConv2D(nn.Module):
    def __init__(self,
                 input_features,
                 output_features,
                 filter_size,
                 padding,
                 stride,
                 num_rotations,
                 upsampling,
                 first_layer):

        super().__init__()
        self.input_features = input_features
        self.output_features = output_features
        self.padding = padding
        self.stride = stride
        self.upsampling = upsampling
        self.n_coeffs = filter_size * (filter_size + 1) // 2

        coeffs = nn.Parameter(
            torch.Tensor(
                self.n_coeffs, self.input_features, self.output_features
            )
        )
        self.coeffs = coeffs

        self.rotate_hermite = RotateHermite(
            filter_size=filter_size,
            upsampling=upsampling,
            num_rotations=num_rotations,
            first_layer=first_layer
        )

        self.weights_all_rotations = None

    def forward(self, input):
        weights_all_rotations = self.rotate_hermite(self.coeffs)
        weights_all_rotations = downsample_weights(weights_all_rotations, self.upsampling)
        weights_all_rotations = weights_all_rotations.permute(3, 2, 0, 1)
        self.weights_all_rotations = weights_all_rotations

        return nn.functional.conv2d(
            input=input,
            weight=self.weights_all_rotations,
            bias=None,
            stride=self.stride,
            padding=self.padding
        )


class RotationEquivariantBatchNorm2D(nn.Module):
    def __init__(self,
                 num_features,
                 num_rotations,
                 eps=1e-05,
                 momentum=0.1,
                 affine=True,
                 track_running_stats=True):

        super().__init__()

        self.num_features = num_features
        self.num_rotations = num_rotations
        self.eps = eps
        self.affine = affine
        self.track_running_stats = track_running_stats
        self.batch_norm = nn.BatchNorm1d(
            num_features=num_features,
            eps=eps,
            momentum=momentum,
            affine=affine,
            track_running_stats=track_running_stats
        )

    def forward(self, input):
        s = input.shape
        input = self.reshape(input, s)
        output = self.batch_norm(input)
        output = self.inv_reshape(output, s)
        return output

    def reshape(self, x, s):
        x = x.view(s[0], self.num_rotations, self.num_features, s[2], s[3])
        x = x.permute(0, 2, 1, 3, 4)
        x = x.reshape(s[0], self.num_features, self.num_rotations * s[2] * s[3])
        return x

    def inv_reshape(self, x, s):
        x = x.view(s[0], self.num_features, self.num_rotations, s[2], s[3])
        x = x.permute(0, 2, 1, 3, 4)
        x = x.reshape(s[0], self.num_rotations * self.num_features, s[2], s[3])
        return x


class RotationEquivariantBias2DLayer(nn.Module):
    def __init__(self, channels, num_rotations, initial=0, **kwargs):
        super().__init__(**kwargs)

        self.num_features = channels
        self.num_rotations = num_rotations
        self.bias = torch.nn.Parameter(torch.empty((1, 1, channels, 1, 1)).fill_(initial))

    def forward(self, x):
        s = x.shape
        x = x.view(s[0], self.num_rotations, self.num_features, s[2], s[3])
        x = x + self.bias
        return x.view(*s)


class RotationEquivariantScale2DLayer(nn.Module):
    def __init__(self, channels, num_rotations, initial=1, **kwargs):
        super().__init__(**kwargs)

        self.num_features = channels
        self.num_rotations = num_rotations
        self.scale = torch.nn.Parameter(torch.empty((1, 1, channels, 1, 1)).fill_(initial))

    def forward(self, x):
        s = x.shape
        x = x.view(s[0], self.num_rotations, self.num_features, s[2], s[3])
        x = x * self.scale
        return x.view(*s)



from mlutils.utils import flatten_json
import pytest


def nested_dictionary(duplicates):
    dictionary = dict(a=0, dict2=dict(b=0, dict3=dict(c=0)))
    if duplicates:
        dictionary["c"] = 0
    return dictionary


def flat_dictionary(keep_nested_name):
    if keep_nested_name:
        return dict(a=0, dict2_b=0, dict2_dict3_c=0)
    else:
        return dict(a=0, b=0, c=0)


@pytest.mark.parametrize("keep_nested_name", [True, False])
def test_output(keep_nested_name):
    assert flatten_json(nested_dictionary(duplicates=False), keep_nested_name) == flat_dictionary(keep_nested_name)


def test_exception():
    with pytest.raises(ValueError):
        flatten_json(nested_dictionary(duplicates=True), keep_nested_name=False)



import torch

from torch import nn


class Bias2DLayer(nn.Module):
    def __init__(self, channels, initial=0, **kwargs):
        super(Bias2DLayer, self).__init__(**kwargs)

        self.bias = torch.nn.Parameter(torch.empty((1, channels, 1, 1)).fill_(initial))

    def forward(self, x):
        return x + self.bias


class Scale2DLayer(nn.Module):
    def __init__(self, channels, initial=1, **kwargs):
        super(Scale2DLayer, self).__init__(**kwargs)

        self.scale = torch.nn.Parameter(torch.empty((1, channels, 1, 1)).fill_(initial))

    def forward(self, x):
        return x * self.scale
